import networkx as nx
from binny.utils import *


def load_hmmer_profiles(hmmer_file):
    profiles = {}
    current_profile = ''
    with open(hmmer_file, 'r') as f:
        for line in f:
            current_profile += line
            if line.startswith('ACC   '):
                profile_name = line.split()[-1].strip('\n').split('.')[0]
            if line.startswith('//'):
                profiles[profile_name] = current_profile
                current_profile = ''
    return profiles


def get_all_marker_set_markers(markers_sets_file):
    all_markers = set()
    with open(markers_sets_file, 'r') as f:
        for line in f:
            line = line.strip('\n \t').split('\t')
            marker_sets = line[-1].replace(']), set([', ';')
            for char in '[](\'() \"':
                marker_sets = marker_sets.replace(char, '').replace('set', '')
            marker_sets = [[marker.split('.')[0] for marker in marker_set.split(',')] for marker_set in
                           marker_sets.split(';') if not marker_set.split(',') == ['']]
            marker_sets = [marker for marker_set in marker_sets for marker in marker_set]
            all_markers.update(marker_sets)
    return all_markers


def get_tigrfam2pfam_dict(tigrfam2pfam_file):
    tigrfam2pfam_dict = {}
    with open(tigrfam2pfam_file, 'r') as f:
        for line in f:
            line = line.strip('\n \t').split('\t')
            tigrfam, pfam = line[1], line[0].replace('pfam', 'PF')
            if not tigrfam2pfam_dict.get(tigrfam):
                tigrfam2pfam_dict[tigrfam] = [pfam]
            else:
                tigrfam2pfam_dict[tigrfam].append(pfam)
            if not tigrfam2pfam_dict.get(pfam):
                tigrfam2pfam_dict[pfam] = [tigrfam]
            else:
                tigrfam2pfam_dict[pfam].append(tigrfam)
    return tigrfam2pfam_dict


def remove_unused_checkm_hmm_profiles(hmmer_file, markers_sets_file, tigrfam2pfam_file, out_path):
    profiles = load_hmmer_profiles(hmmer_file)
    profiles_accs = list(profiles.keys())
    all_markers = get_all_marker_set_markers(markers_sets_file)
    tf2pf = get_tigrfam2pfam_dict(tigrfam2pfam_file)
    out_path_pfam = Path(out_path + '/checkm_pf/checkm_filtered_pf.hmm')
    out_path_tigrfam = Path(out_path + '/checkm_tf/checkm_filtered_tf.hmm')
    out_path_pfam.parent.mkdir(parents=True, exist_ok=True)
    out_path_tigrfam.parent.mkdir(parents=True, exist_ok=True)
    with open(out_path_pfam, 'w') as of_pfam:
        with open(out_path_tigrfam, 'w') as of_tigrfam:
            for profile in profiles_accs:
                if profile in all_markers or any(alt_acc in all_markers for alt_acc in tf2pf.get(profile, [])):
                    if profile.startswith('PF'):
                        of_pfam.write(profiles[profile])
                    elif profile.startswith('TIGR'):
                        of_tigrfam.write(profiles[profile])


def load_checkm_markers(marker_file):
    linegae_dict = {0: 'domain', 1: 'phylum', 2: 'class', 3: 'order', 4: 'family', 5: 'genus', 6: 'species'}
    tms_data = nx.DiGraph()
    with open(marker_file, 'r') as f:
        for line in f:
            line = line.strip('\n \t').split('\t')
            if (int(line[5]) < 58
                    or 'Tenericutes' in line[2]
                    or 'Haemophilus' in line[2]
                    # or 'Enterobacteriales' in line[2]
                    or 'Rickettsiales' in line[2]
                    or 'Chlamydiae' in line[2]):
                logging.debug('Skipping loading of markers from {0}.'.format(line[1]))
                continue
            marker_sets = eval(line[-1])
            marker_sets = [{marker.split('.')[0] for marker in marker_set} for marker_set in
                           marker_sets]
            lineage = line[2].split(';')
            lineage = [tax + '[' + linegae_dict[tax_ind] + ']' if tax == lineage[tax_ind - 1] and len(lineage) > 1
                       else tax for tax_ind, tax in enumerate(lineage)]
            if len(lineage) == 7:
                lineage[-1] = ' '.join(lineage[-2:])

            if len(marker_sets) >= 10:
                tms_data.add_nodes_from(lineage)
                tms_data.add_edges_from([(i, lineage[index + 1]) for index, i in enumerate(lineage)
                                         if not index + 1 == len(lineage)])
                tms_data.nodes[lineage[-1]]['markers'] = int(line[4])
                tms_data.nodes[lineage[-1]]['marker_groups'] = int(line[5])
                tms_data.nodes[lineage[-1]]['marker_sets'] = marker_sets

                # Check if higher level tax marker sets are identical
                if not len(lineage) == 1:
                    for level in lineage[:-1]:
                        try:
                            if tms_data.nodes[level]['marker_sets'] == marker_sets:
                                tms_data.nodes[lineage[-1]]['marker_sets'] = 'is_' + level
                                break
                        except KeyError:
                            logging.warning(lineage)
                            logging.warning(tms_data.nodes[lineage[-1]])
                            raise KeyError
    return tms_data


def checkm_hmmer_search2prokka_gff(hmm_checkm_marker_out, prokka_gff, gff_out_path='intermediary'):
    prokka_checkm_gff = '{0}/annotation_CDS_RNA_hmms_checkm.gff'.format(gff_out_path)
    checkm_marker_dict = parse_mantis_cons_annot(hmm_checkm_marker_out)
    with open(prokka_checkm_gff, 'w') as pcg:
        with open(prokka_gff, 'r') as pg:
            for line in pg:
                line = line.strip('\n \t')
                line_annots = line.split('\t')[-1].split(';')
                line_gene = line_annots[0].replace('ID=', '')
                if checkm_marker_dict.get(line_gene):
                    pcg.write(line + ';checkm_marker=' + ','.join(checkm_marker_dict[line_gene]) + '\n')
                else:
                    pcg.write(line + '\n')


if __name__ == '__main__':
    pass